//+------------------------------------------------------------------+
//|                                    candles.mq4                   |
//|                                    www.globalmarketingdesign.com |
//+------------------------------------------------------------------+
#property copyright ""
#property link      "www.globalmarketingdesign.com"
/*
$Log: /TEXASNOMAD/candles.mq4 $
 * 
 * 1     1/03/09 22:43 Pion
 * first version with reliable orders
 * 
*/

#include <stderror.mqh>

//Completion to stderror.mqh
#define           ERR_TRADE_EXPIRATION_DENIED	147	//Expirations are denied by broker.
#define           OBJECT_NAME_PATTERN           "c4n8l3s"   

//+------------------------------------------------------------------+
//| Inputs                                                           |
//+------------------------------------------------------------------+
extern string     Header__1                     = " --------------- Initial Parameters ------------------ ";
extern double     LotSize                       = 0.5;
extern double     TailHeight                    = 10;
extern int        MinCandleHeight               = 5;
extern int        NoCoverRule                   = 0;
extern int        MomentumRule1                 = 0;
extern int        MomentumRule2                 = 0;
extern int        TakeProfit                    = 20;
extern int        StopLoss                      = 10;
extern int        TrailingStop                  = 10;
extern int        ActiveHours                   = 5;
extern int        barCount                      = 10;
 
//+------------------------------------------------------------------+
//| Globals                                                          |
//+------------------------------------------------------------------+
int MagicNumber;

//+------------------------------------------------------------------+
//| expert initialization function                                   |
//+------------------------------------------------------------------+
int init()
{
   MagicNumber = WindowHandle(Symbol(), Period());

   DeleteAllObjects();
   DrawLabels("www.globalmarketingdesign.com", 1); 
   
   return(0);
}

//+------------------------------------------------------------------+
//| expert deinitialization function                                 |
//+------------------------------------------------------------------+
int deinit()
{
   DeleteAllObjects();
   
   return(0);
}

//+------------------------------------------------------------------+
//| expert start function                                            |
//+------------------------------------------------------------------+
int start()
{  
   double Lo = Low[0];
   double Hi = High[0];
   double Op = Open[0];
   double Cl = Close[0];
   
   bool longTradeOpen = false;
   bool shortTradeOpen = false;
   bool longTradeClose = false;
   bool shortTradeClose = false;
   
   double diff;
   int count;
  
   // Long Open
   if(CheckLongOpen(Lo, Hi, Op, Cl) >= TailHeight) {
      longTradeOpen = true;
   }
   
   for(count = 0; count < MomentumRule2; count++) {
     diff = Low[count];
     
     if(diff > MomentumRule1) {
      longTradeOpen = true;
     }
   }
   
   if(iLowest(NULL, 0, MODE_LOW, barCount, barCount) < TrailingStop) {
      longTradeClose = true;
   }
   
   // Long Close
   if(CheckLongClose(Lo, Hi, Op, Cl) >= TailHeight) {
      longTradeClose = true;
   }
  
   // Short Open
   if(CheckShortOpen(Lo, Hi, Op, Cl) >= TailHeight) {
      shortTradeOpen = true;
   }
   
   for(count = 0; count < MomentumRule2; count++) {
     diff = High[count];
     
     if(diff > MomentumRule1) {
      shortTradeOpen = true;
     }
   }
   
   if(iHighest(NULL, 0, MODE_HIGH, barCount, barCount) > TrailingStop) {
      shortTradeClose = true;
   }
   
   // Short Close
   if(CheckShortClose(Lo, Hi, Op, Cl) >= TailHeight) {
      shortTradeClose = true;
   }
   
   // Trade
   if(longTradeOpen == true && CountLongs(MagicNumber)== 0) {
      CloseShorts(MagicNumber);
      ReliableOrderSend(Symbol(),OP_BUY,LotSize,Ask,TrailingStop,StopLong(Bid,StopLoss),TakeLong(Ask,TakeProfit),"bb10466",MagicNumber,0,Green,10); 
   }
   
   if(shortTradeOpen == true && CountLongs(MagicNumber)== 0)
   {
      CloseLongs(MagicNumber);
      ReliableOrderSend(Symbol(),OP_SELL,LotSize,Bid,TrailingStop,StopShort(Ask,StopLoss),TakeShort(Bid,TakeProfit),"bb10466",MagicNumber,0,Red,10); 
   } 
   
   if(longTradeClose == true && CountLongs(MagicNumber)== 1) {
      CloseLongs(MagicNumber);
   }
   
   if(shortTradeClose == true && CountLongs(MagicNumber)== 1)
   {
      CloseShorts(MagicNumber); 
   } 
   
   return(0);
}

//+------------------------------------------------------------------+
//| Check Long Open                                                 |
//+------------------------------------------------------------------+
double CheckLongOpen(double Lo, double Hi, double Op, double Cl)
{     
   // Long Trade, Open Candle: 
   // Open - Low
   // ---------- = n%
   // High - Low
   return (((Op-Lo) / (Hi-Lo)) * 100);
}

//+------------------------------------------------------------------+
//| Check Long Close                                                 |
//+------------------------------------------------------------------+
double CheckLongClose(double Lo, double Hi, double Op, double Cl)
{     
   // Long Trade, Closed Candle:
   // Close - Low
   // ----------- = n%
   // High - Low 
   return (((Cl-Lo) / (Hi-Lo)) * 100);
}

//+------------------------------------------------------------------+
//| Check Short Open                                                 |
//+------------------------------------------------------------------+
double CheckShortOpen(double Lo, double Hi, double Op, double Cl)
{     
   // Short Trade, Open Candle:
   // High - Close
   // ------------ = n%
   // High - Low
   return (((Hi-Cl) / (Hi-Lo)) * 100);
}

//+------------------------------------------------------------------+
//| Check Short Close                                                |
//+------------------------------------------------------------------+
double CheckShortClose(double Lo, double Hi, double Op, double Cl)
{     
   // Short Trade, Closed Candle:
   // High - Open
   // ----------- = n%
   // High - Low
   return (((Hi-Op) / (Hi-Lo)) * 100);
}

//+------------------------------------------------------------------+
//| Calculate concurrent Long position                               |
//+------------------------------------------------------------------+
int CountLongs(int MagicNumber)
{
 int count=0;
 int trade;
 int trades=OrdersTotal();
 for(trade=0;trade<trades;trade++)
 {
  OrderSelect(trade,SELECT_BY_POS,MODE_TRADES);
  
  if( OrderSymbol()!=Symbol() || OrderMagicNumber() != MagicNumber )
   continue;
   
  if(OrderType()==OP_BUY)
   count++;
 }//for
 return(count);
}

//+------------------------------------------------------------------+
//| Calculate concurrent short position                              |
//+------------------------------------------------------------------+
int CountShorts(int MagicNumber)
{
 int count=0;
 int trade;
 int trades=OrdersTotal();
 for(trade=0;trade<trades;trade++)
 {
  OrderSelect(trade,SELECT_BY_POS,MODE_TRADES);
  
  if(OrderSymbol()!=Symbol() || OrderMagicNumber() != MagicNumber )
   continue;
   
  if(OrderType()==OP_SELL)
  count++;
 }//for
 return(count);
}

//+------------------------------------------------------------------+
//| Calculate Stop Long                                              |
//+------------------------------------------------------------------+
double StopLong(double price,int stop)
{
 if(stop==0) {  return(0); }
 else        {  return(price-(stop*Point));}
}

//+------------------------------------------------------------------+
//| Calculate Stop Short                                             |
//+------------------------------------------------------------------+
double StopShort(double price,int stop)
{
 if(stop==0) {    return(0);}
 else        {    return(price+(stop*Point));}
}

//+------------------------------------------------------------------+
//| Close Long Position                                              |
//+------------------------------------------------------------------+
void CloseLongs(int MagicNumber)
{
 int i = OrdersTotal();
 
 while( CountLongs(MagicNumber) != 0 && i >= 0 )
 {  
   OrderSelect(i,SELECT_BY_POS,MODE_TRADES);
   
   if( OrderSymbol() != Symbol() || OrderMagicNumber() != MagicNumber )
   {
     i--;
     continue;
   }
   else if(OrderType()==OP_BUY || OrderType()== OP_BUYLIMIT)
   {
     ReliableOrderClose(OrderTicket(), OrderLots(), Ask, TrailingStop, Blue, 10);
     i--;
   }
 }
}

//+------------------------------------------------------------------+
//| Close Short Position                                             |
//+------------------------------------------------------------------+
void CloseShorts(int MagicNumber)
{
 int i = OrdersTotal();
 
 while( CountShorts(MagicNumber) != 0 && i >= 0 )
 {  
   OrderSelect(i,SELECT_BY_POS,MODE_TRADES);
   
   if( OrderSymbol() != Symbol() || OrderMagicNumber() != MagicNumber)
   {
     i--;
     continue;
   }
   else if(OrderType()== OP_SELL || OrderType()==OP_SELLLIMIT )
   {
     ReliableOrderClose(OrderTicket(), OrderLots(), Ask, TrailingStop, Pink, 10);
     i--; 
   }
 }
}

//+------------------------------------------------------------------+
//| Calculate Profit Target Long                                     |
//+------------------------------------------------------------------+
double TakeLong(double price,int take)
{
 if(take==0) {  return(0);}
 else        {  return(price+(take*Point));}
}

//+------------------------------------------------------------------+
//| Calculate Profit Target Short                                     |
//+------------------------------------------------------------------+
double TakeShort(double price,int take)
{
 if(take==0) {  return(0);}
 else        {  return(price-(take*Point));}
}

//+------------------------------------------------------------------+
//| Delete Objects                                                   |
//+------------------------------------------------------------------+
void DeleteAllObjects()
{     
   int objectCount = ObjectsTotal();
   for (int index = objectCount - 1; index >= 0; index--)
   {
      string objectName = ObjectName(index);
      if (StringSubstr(objectName, 0, StringLen(OBJECT_NAME_PATTERN)) == OBJECT_NAME_PATTERN)
         ObjectDelete(objectName);
   }   
}

//+------------------------------------------------------------------+
//| Draw Copywrite                                                   |
//+------------------------------------------------------------------+ 
double DrawLabels(string text, int lineIndex)
{
   string objName = OBJECT_NAME_PATTERN + lineIndex + "_Copyright";
   int xPos = 8;
   int yPos = lineIndex * 15;
   int cornerID = 1;
   ObjectDelete(objName);
   ObjectCreate(objName, OBJ_LABEL, 0, 0, 0);
   ObjectSet(objName, OBJPROP_CORNER, cornerID);
   ObjectSet(objName, OBJPROP_XDISTANCE, xPos);
   ObjectSet(objName, OBJPROP_YDISTANCE, yPos);
   ObjectSet(objName, OBJPROP_BACK, true);
   ObjectSetText(objName, text, 14, "Times", Orange);
   WindowRedraw();
}

//+------------------------------------------------------------------+
//| Price Open Mode                                                  |
//+------------------------------------------------------------------+ 
int PriceOpenMode(int op)
  {
  if (op==OP_BUY||op==OP_BUYLIMIT||op==OP_BUYSTOP)
     return(MODE_ASK);
  if (op==OP_SELL||op==OP_SELLLIMIT||op==OP_SELLSTOP)
     return(MODE_BID);
  return(-1);
}

//+------------------------------------------------------------------+
//| Price Close Mode                                                 |
//+------------------------------------------------------------------+ 
int PriceCloseMode(int op)
  {
  if (op==OP_BUY||op==OP_BUYLIMIT||op==OP_BUYSTOP)
     return(MODE_BID);
  if (op==OP_SELL||op==OP_SELLLIMIT||op==OP_SELLSTOP)
     return(MODE_ASK);
  return(-1);
}

//+------------------------------------------------------------------+
//| Open Reliable Order                                              |
//+------------------------------------------------------------------+ 
int ReliableOrderSend(string symbol,int cmd,double volume,double price,int slippage,double stoploss,double takeprofit, string comment="",int magic=0,datetime expiration=0,color arrow_color=CLR_NONE,int MaxPasses=0) 
  {  
  int Gle=ERR_TRADE_CONTEXT_BUSY;    
  string ErrType="";
  string ErrText="";
  string ErrExplanation="";    
  int passes=0;  
  int res=-1;  
  while (Gle==ERR_TRADE_CONTEXT_BUSY||Gle==ERR_REQUOTE||Gle==ERR_INVALID_PRICE||Gle==ERR_PRICE_CHANGED||Gle==ERR_OFF_QUOTES)
       {  
          
       if (Gle==ERR_REQUOTE||Gle==ERR_INVALID_PRICE||Gle==ERR_PRICE_CHANGED||Gle==ERR_OFF_QUOTES||passes==0)
         {
         if (passes!=0)
            RefreshRates();
         if (price==0.0)  //if (passes!=0||price==0)
            price=MarketInfo(symbol,PriceOpenMode(cmd));
         }//if (Gle==ERR_REQUOTE)                        
       res=OrderSend(symbol,cmd,volume,price,slippage,stoploss,takeprofit,comment,magic,expiration,arrow_color);   
       
       Gle=GetLastError();
       TranslateError(Gle,ErrType,ErrText,ErrExplanation);     
       
       if (Gle!=ERR_NO_ERROR)
          Print("ReliableOrderSend error : ",Gle," : ",ErrText);              
       passes=passes+1;
       
       if (MaxPasses!=0)
         {
          if (passes>=MaxPasses)
            break;
         }

       if (Gle==ERR_REQUOTE||Gle==ERR_INVALID_PRICE||Gle==ERR_PRICE_CHANGED||Gle==ERR_OFF_QUOTES) { price=0.0; }
         
       }//while (Gle==ERR_TRADE_CONTEXT_BUSY||Gle==ERR_REQUOTE)       
  return(res);
 }
     
//+------------------------------------------------------------------+
//| Close Reliable Order                                             |
//+------------------------------------------------------------------+ 
bool ReliableOrderClose(int ticket, double lots, double price, int slippage, color Color=CLR_NONE,int MaxPasses=0) 
  {
  int Gle=ERR_TRADE_CONTEXT_BUSY;
  string ErrType="";
  string ErrText="";
  string ErrExplanation="";    
  int passes=0;  
  bool res;
  int otype;
  double olots;
  string osymbol;
  res=OrderSelect(ticket,SELECT_BY_TICKET,MODE_TRADES);
  osymbol=OrderSymbol();
  otype=OrderType();
  olots=OrderLots();  
  if (lots==0)
     lots=olots;  
  if (res==True)
    {
    while (Gle==ERR_TRADE_CONTEXT_BUSY||Gle==ERR_REQUOTE||Gle==ERR_INVALID_PRICE||Gle==ERR_PRICE_CHANGED||Gle==ERR_OFF_QUOTES)
       {     
       if (Gle==ERR_REQUOTE||Gle==ERR_INVALID_PRICE||Gle==ERR_PRICE_CHANGED||Gle==ERR_OFF_QUOTES||passes==0)
         {
         if (passes!=0)
            RefreshRates();
         if (price==0.0)  //if (passes!=0||price==0)
            price=MarketInfo(osymbol,PriceCloseMode(otype));
         }//if (Gle==ERR_REQUOTE)                               
       res=OrderClose(ticket,lots,price,slippage,Color);  
        
       Gle=GetLastError();
       TranslateError(Gle,ErrType,ErrText,ErrExplanation);
       
       if (Gle!=ERR_NO_ERROR)
          Print("ReliableOrderClose error : ",Gle," : ",ErrText);           
       passes=passes+1;
       
       if (MaxPasses!=0)
         {
          if (passes>=MaxPasses)
            break;
         }
       
       if (Gle==ERR_REQUOTE||Gle==ERR_INVALID_PRICE||Gle==ERR_PRICE_CHANGED||Gle==ERR_OFF_QUOTES) { price=0.0; }

       }//while (Gle==ERR_TRADE_CONTEXT_BUSY||Gle==ERR_REQUOTE)  
    }
  return(res);
 }
 
//+------------------------------------------------------------------+
//| Translate Order                                                  |
//+------------------------------------------------------------------+ 
 void TranslateError(int errcode,string &ErrType,string &ErrText,string &ErrExplanation)
     {
     ErrType="Unknown type.";
     ErrText="Unknown error.";
     ErrExplanation="Unknown reason";                
     //no need to inquire why damn switch is not working
      if (errcode==ERR_NO_ERROR) 
         {
         ErrType="Server error.";
         ErrText="No error.";
         ErrExplanation="Trade operation succeeded."; 
         return;
         }         
      if (errcode==ERR_NO_RESULT) 
         {
         ErrType="Server error.";
         ErrText="No result.";
         ErrExplanation="OrderModify attempts to replace the values already set with the same values. One or more values must be changed, then modification attempt can be repeated.";
         return;
         }                  
      if (errcode==ERR_COMMON_ERROR) 
         {
         ErrType="Server error.";
         ErrText="Common error.";
         ErrExplanation="Common error. All attempts to trade must be stopped until reasons are clarified. Restart of operation system and client terminal will possibly be needed.";  
         return;
         }         
      if (errcode==ERR_INVALID_TRADE_PARAMETERS) 
         {
         ErrType="Server error.";
         ErrText="Invalid trade parameters.";
         ErrExplanation="Invalid parameters were passed to the trading function, for example, wrong symbol, unknown trade operation, negative slippage, non-existing ticket number, etc. The program logic must be changed."; 
         return;
         }    
      if (errcode==ERR_SERVER_BUSY) 
         {
         ErrType="Server error.";
         ErrText="Server busy.";
         ErrExplanation="Trade server is busy. The attempt can be repeated after a rather long period of time (over several minutes)."; 
         return;
         }         
      if (errcode==ERR_OLD_VERSION) 
         {
         ErrType="Server error.";
         ErrText="Old version.";
         ErrExplanation="Old version of the client terminal. The latest version of the client terminal must be installed."; 
         return;
         }
      if (errcode==ERR_NO_CONNECTION) 
         {
         ErrType="Server error."; 
         ErrText="No connection.";
         ErrExplanation="No connection to the trade server. It is necessary to make sure that connection has not been broken (for example, using the IsConnected function) and repeat the attempt after a certain period of time (over 5 seconds)."; 
         return;
         }
      if (errcode==ERR_TOO_FREQUENT_REQUESTS) 
         {
         ErrType="Server error.";
         ErrText="Too frequent request.";
         ErrExplanation="Requests are too frequent. The frequency of requesting must be reduced, the program logic must be changed."; 
         return;
         }    
      if (errcode==ERR_ACCOUNT_DISABLED) 
         {
         ErrType="Server error.";
         ErrText="Account disabled.";
         ErrExplanation="The account was disabled. All attempts to trade must be stopped."; 
         return;
         }
      if (errcode==ERR_INVALID_ACCOUNT) 
         {
         ErrType="Server error.";
         ErrText="Invalid account.";
         ErrExplanation="The account number is invalid. All attempts to trade must be stopped."; 
         return;
         }     
      if (errcode==ERR_TRADE_TIMEOUT) 
         {
         ErrType="Server error.";
         ErrText="Trade timeout.";
         ErrExplanation="Timeout for the trade has been reached. Before retry (at least, in 1-minute time),"+
         " it is necessary to make sure that trading operation has not really succeeded (a new position has not been opened, or the existing order has not been modified or deleted, or the existing position has not been closed)."; 
         return;         
         }
      if (errcode==ERR_INVALID_PRICE) 
         {
         ErrType="Server error.";
         ErrText="Invalid price.";
         ErrExplanation="Invalid bid or ask price, perhaps, unnormalized price. After 5-second (or more) "+
         "delay, it is necessary to refresh data using the RefreshRates function and make a retry. If the error does not disappear, all attempts to trade must be stopped, the program logic must be changed."; 
         return;
         }
      if (errcode==ERR_INVALID_STOPS) 
         {
         ErrType="Server error.";
         ErrText="Invalid stops.";
         ErrExplanation="Stops are too close, or prices are ill-calculated or unnormalized (or in the open "+
         "price of a pending order). The attempt can be repeated only if the error occurred due to the price "+
         "obsolescense. After 5-second (or more) delay, it is necessary to refresh data using the RefreshRates function and make a retry. If the error does not disappear, all attempts to trade must be stopped, the program logic must be changed.";
         return;
         }         
      if (errcode==ERR_INVALID_TRADE_VOLUME) 
         {
         ErrType="Server error.";
         ErrText="Invalid trade volume:";
         ErrExplanation="Invalid trade volume, error in the volume granularity. All attempts to trade must be stopped, and the program logic must be changed."; 
         return;         
         }
      if (errcode==ERR_MARKET_CLOSED) 
         {
         ErrType="Server error.";
         ErrText="Market closed.";
         ErrExplanation="Market is closed. The attempt can be repeated after a rather long period of time (over several minutes)."; 
         return;
         }     
      if (errcode==ERR_TRADE_DISABLED) 
         {
         ErrType="Server error.";
         ErrText="Trade disabled.";
         ErrExplanation="Trade is disabled. All attempts to trade must be stopped."; 
         return;         
         }
      if (errcode==ERR_NOT_ENOUGH_MONEY) 
         {      
         ErrType="Server error.";
         ErrText="Not enough money.";
         ErrExplanation="Not enough money to make an operation. The trade with the same parameters must not be repeated."+         
         " After 5-second (or more) delay, the attempt can be repeated with a smaller volume, but it is necessary to make sure that there is enough money to complete the operation."; 
         return;
         }         
      if (errcode==ERR_PRICE_CHANGED) 
         {
         ErrType="Server error.";
         ErrText="Price changed.";
         ErrExplanation="The price has changed. The data can be refreshed without any delay using the RefreshRates function and make a retry."; 
         return;
         }         
      if (errcode==ERR_OFF_QUOTES) 
         {
         ErrType="Server error.";
         ErrText="No quotes.";
         ErrExplanation="No quotes. The broker has not supplied with prices or refused, for any reason (for example, no prices"+
         " at the session start, unconfirmed prices, fast market). After 5-second (or more) delay, it is necessary to refresh data using the RefreshRates function and make a retry."; 
         return;
         }         
      if (errcode==ERR_REQUOTE) 
         {         
         ErrType="Server error.";
         ErrText="Requote.";
         ErrExplanation="The requested price has become out of date (more than 5 ticks have income) or bid and ask prices"+
         " have been mixed up. The data can be refreshed without any delay using the RefreshRates function and make a retry. If the error does not disappear, all attempts to trade must be stopped, the program logic must be changed."; 
         return;
         }         
      if (errcode==ERR_ORDER_LOCKED) 
         {
         ErrType="Server error.";
         ErrText="Order locked.";
         ErrExplanation="The order has been locked and under processing. All attempts to make trading operations must be stopped, and the program logic must be changed."; 
         return;
         }         
      if (errcode==ERR_LONG_POSITIONS_ONLY_ALLOWED) 
         {
         ErrType="Server error.";
         ErrText="Long positions only allowed.";
         ErrExplanation="Only buying operation is allowed. The SELL operation must not be repeated."; 
         return;         
         }
      if (errcode==ERR_TOO_MANY_REQUESTS) 
         {
         ErrType="Server error.";
         ErrText="Too many many requests.";
         ErrExplanation="Too many requests. The frequency of requesting must be reduced, the program logic must be changed."; 
         return;
         }         
      if (errcode==ERR_TRADE_MODIFY_DENIED) 
         {
         ErrType="Server error.";
         ErrText="Trade modify denied.";
         ErrExplanation="Modifying has been denied since the order is too close to market and locked for possible soon execution. The data can be refreshed after more than 15 seconds using the RefreshRates function, and a retry can be made."; 
         return;
         }         
      if (errcode==ERR_TRADE_CONTEXT_BUSY) 
         {
         ErrType="Server error.";
         ErrText="Trade context busy.";
         ErrExplanation="The trade thread is busy. Retry only after the IsTradeAllowed function has returned TRUE."; 
         return;
         }         
      if (errcode==ERR_TRADE_EXPIRATION_DENIED) 
         {
         ErrType="Server error";
         ErrText="Trade expiration denied.";
         ErrExplanation="The use of pending order expiration date has been denied by the broker. The operation can only be repeated if the expiration parameter has been zeroized."; 
         return;
         }         
      if (errcode>=ERR_NO_MQLERROR)
         {
         ErrType="Runtime error.";
         ErrExplanation="Not available.";
         if (errcode==ERR_NO_MQLERROR)
            {
            ErrText="No error.";
            ErrExplanation="Not needed.";
            return;
            }
         if (errcode==4001)
            {
            ErrText="Wrong function pointer.";
            return;
            }          
         if (errcode==4002)
            {
            ErrText="Array index is out of range.";
            return;
            }          
         if (errcode==4003)
            {
            ErrText="No memory for function call stack.";
            return;
            }          
         if (errcode==4004)
            {
            ErrText="Recursive stack overflow.";
            return;
            }          
         if (errcode==4005)
            {
            ErrText="Not enough stack for parameter.";
            return;
            }          
         if (errcode==4006)
            {
            ErrText="No memory for parameter string.";
            return;
            }          
         if (errcode==4007)
            {
            ErrText="No memory for temp string.";
            return;
            }          
         if (errcode==4008)
            {
            ErrText="Not initialized string.";
            return;
            }          
         if (errcode==4009)
            {
            ErrText="Not initialized string in array.";
            return;
            }      
         if (errcode==4010)
            {
            ErrText="No memory for array string.";
            return;
            }
         if (errcode==4011)
            {
            ErrText="Too long string.";
            return;
            }          
         if (errcode==4012)
            {
            ErrText="Remainder from zero divide.";
            return;
            }          
         if (errcode==4013)
            {
            ErrText="Zero divide.";
            return;
            }          
         if (errcode==4014)
            {
            ErrText="Unknown command.";
            return;
            }          
         if (errcode==4015)
            {
            ErrText="Wrong jump (never generated error).";
            return;
            }          
         if (errcode==4016)
            {
            ErrText="Not initialized array.";
            return;
            }          
         if (errcode==4017)
            {
            ErrText="DLL calls are not allowed.";
            return;
            }          
         if (errcode==4018)
            {
            ErrText="Cannot load library.";
            return;
            }          
         if (errcode==4019)
            {
            ErrText="Cannot call function.";
            return;
            }                      
         if (errcode==4020)
            {
            ErrText="Expert function calls are not allowed.";
            return;
            }
         if (errcode==4021)
            {
            ErrText="Not enough memory for temp string returned from function.";
            return;
            }          
         if (errcode==4022)
            {
            ErrText="System is busy (never generated error).";
            return;
            }
         if (errcode==4050)
            {
            ErrText="Invalid function for parameters count.";
            return;
            }
         if (errcode==4051)
            {
            ErrText="Invalid function parameter value.";
            return;
            }          
         if (errcode==4052)
            {
            ErrText="String function internal error.";
            return;
            }          
         if (errcode==4053)
            {
            ErrText="Some array error.";
            return;
            }          
         if (errcode==4054)
            {
            ErrText="Incorrect series array using.";            
            return;
            }          
         if (errcode==4055)
            {
            ErrText="Custom indicator error.";
            return;
            }          
         if (errcode==4056)
            {
            ErrText="Global variables processing error.";
            return;
            }          
         if (errcode==4057)
            {
            ErrText="Global variables processing error.";
            return;
            }          
         if (errcode==4058)
            {
            ErrText="Global variable not found.";
            return;
            }          
         if (errcode==4059)
            {
            ErrText="Function is not allowed in testing mode."; 
            return;
            }               
         if (errcode==4060)
            {
            ErrText="Function is not confirmed.";
            return;
            }
         if (errcode==4061)
            {
            ErrText="Send mail error.";
            return;
            }          
         if (errcode==4062)
            {
            ErrText="String parameter expected.";
            return;
            }          
         if (errcode==4063)
            {
            ErrText="Integer parameter expected.";
            return;
            }          
         if (errcode==4064)
            {
            ErrText="Double parameter expected.";            
            return;
            }          
         if (errcode==4065)
            {
            ErrText="Array as parameter expected.";
            return;
            }          
         if (errcode==4066)
            {
            ErrText="Requested history data in updating state.";
            return;
            }          
         if (errcode==4067)
            {
            ErrText="Some error in trading function.";
            return;
            }               
         if (errcode==4099)
            {
            ErrText="End of file.";
            return;
            }                                    
         if (errcode==4100)
            {
            ErrText="Some file error.";
            return;
            }
         if (errcode==4101)
            {
            ErrText="Wrong file name.";
            return;
            }          
         if (errcode==4102)
            {
            ErrText="Too many opened files.";
            return;
            }          
         if (errcode==4103)
            {
            ErrText="Cannot open file.";
            return;
            }          
         if (errcode==4104)
            {
            ErrText="Incompatible access to a file.";            
            return;
            }          
         if (errcode==4105)
            {
            ErrText="No order selected.";
            return;
            }          
         if (errcode==4106)
            {
            ErrText="Unknown symbol.";
            return;
            }          
         if (errcode==4107)
            {
            ErrText="Invalid price.";
            return;
            }          
         if (errcode==4108)
            {
            ErrText="Invalid ticket.";
            return;
            }          
         if (errcode==4109)
            {
            ErrText="Trade is not allowed."; 
            return;
            }               
         if (errcode==4110)
            {
            ErrText="Longs are not allowed.";
            return;
            }
         if (errcode==4111)
            {
            ErrText="Shorts are not allowed";
            return;
            }                      
         if (errcode==4200)
            {
            ErrText="Object exists already.";
            return;
            }
         if (errcode==4201)
            {
            ErrText="Unknown object property.";
            return;
            }          
         if (errcode==4202)
            {
            ErrText="Object does not exist.";
            return;
            }          
         if (errcode==4203)
            {
            ErrText="Unknown object type.";
            return;
            }          
         if (errcode==4204)
            {
            ErrText="No object name.";            
            return;
            }          
         if (errcode==4205)
            {
            ErrText="Object coordinates error.";
            return;
            }          
         if (errcode==4206)
            {
            ErrText="No specified subwindow.";
            return;
            }          
         if (errcode==4207)
            {
            ErrText="Some error in object function.";
            return;
            }                              
         ErrText="Unknown error.";
         ErrExplanation="Unknown reason.";
         return;          
         }                                      
    return;         
 } //end function